/* ------------------------------------------------------------------------- */
/* adap-phil.c i2c-driver for Philips-style adapters using the parallel port */
/* ------------------------------------------------------------------------- */
/*   Copyright (C) 1995-96 Simon G. Vogl

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		     */
/* ------------------------------------------------------------------------- */
static char rcsid[] = "$Id: $";
/*
 * $Log: $
 *
 */

#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <linux/errno.h>
#include <linux/sched.h>

#include "i2c.h"
#include "i2c_priv.h"

/* ----- global defines ---------------------------------------------------- */
#define DEB(x)		/* should be reasonable open, close &c. 	*/
#define DEB2(x) 	/* low level debugging - very slow 		*/
#define DEBE(x)	x	/* error messages 				*/
#define DEBI(x) 	/* ioctl and its arguments 			*/
#define DEBACK(x)	/* ack failed message				*/

/* debugging - slow down transfer to have a look at the data .. 	*/
/* I use this with two leds&resistors, each one connected to sda,scl 	*/
/* respectively. This makes sure that the algorithm works. Some chips   */
/* might not like this, as they have an internal timeout of some mils	*/
/*
#define SLO_IO      jif=jiffies;while(jiffies<=jif+i2c_table[minor].veryslow);
*/
/* ---- minimum delays for transmission	& a bit more for my long lines	*/
#define WHI  10
#define WLO  10


/* ----- global variables -------------------------------------------------- */

#ifdef SLO_IO
	int jif;
#endif

/* ----- local functions --------------------------------------------------- */

/* ----- level 1: controlling the i2c bus   ----- */

static void scllo(int minor)
{
	outb(I2C_CMASK, CTRL); 
	udelay(WLO);
#ifdef SLO_IO
	SLO_IO
#endif 
	udelay(i2c_table[minor].udelay);
}

static int sclhi(int minor)
{
	int start=jiffies;

	outb(I2C_SCL, CTRL); 
	udelay(WHI);				/* at least minimum time*/
	while ((inb(STAT)& I2C_SCLIN)) {	/* wait till high	*/
		outb(I2C_SCL, CTRL);
		udelay(i2c_table[minor].udelay);
		if (start+i2c_table[minor].timeout<=jiffies){
			DEBE(printk("sclhi failed\n"));
			return -ETIMEDOUT;
		}
	}
#ifdef SLO_IO
	SLO_IO
#endif 
	return 0;
} 

static void sdalo(int minor)
{
	outb(I2C_SDA , DATA);
	udelay(WLO);
#ifdef SLO_IO
	SLO_IO
#endif 
}

static void sdahi(int minor)
{
	outb(I2C_DMASK , DATA);
	udelay(WHI);
#ifdef SLO_IO
	SLO_IO
#endif 
}

#define sdain(minor) ( 0 != ( (inb(STAT)) & I2C_SDA ) )


int i2c_reset(int minor)
{
	outb(I2C_SCL, CTRL); 
	udelay(50);
	outb(I2C_DMASK , DATA);
	udelay(50);
	return 0;
}


static void i2c_start(int minor) 
{
	/* assert: scl, sda are high */
	DEB2(printk("i2c_start.\n"));
	sdalo(minor);
	udelay(WHI);   
	scllo(minor);
	udelay(WLO);   
}


static void i2c_stop(int minor) 
{
	DEB2(printk("i2c_stop.\n"));
	/* assert: scl is low */
	sdalo(minor);
	sclhi(minor); 
	udelay(WHI);    
	sdahi(minor);
	udelay(WHI);
}


static int i2c_outb(int minor, char c)
{
	int i;
	char ack;

	/* assert: scl is low */
	DEB2(printk("i2c_outb:%x\n",c));
	for ( i=7 ; i>=0 ; i-- ) {
		if ( c&(1<<i) ){
			sdahi(minor);
		} else {		
			sdalo(minor);
		}
		if (sclhi(minor)<0) { /* timed out */
			sdahi(minor); /* we don't want to block the net */
			return 0;
		};
		/* arbitrate here: */
		udelay(WHI);
		if ( (c&(1<<i)) && !sdain(minor) ){
			/* we lost the arbitration process -> get lost */
			DEBE(printk("i2c_write: arbitration bailout! \n"));
			return 0;
		}
		scllo(minor);	
	}
	sdahi(minor);
	if (sclhi(minor)<0){ /* timeout */
		return 0;
	};
	/* read ack */
	udelay(WHI);
	ack=!sdain(minor);	/* ack: sda is pulled low ->success.	 */
	DEB2(printk("sdain: 0x%x\n",sdain(minor)));
	scllo(minor);
	return ack;		/* return 1 if ok		 */
	/* assert: scl is low */
}

static char i2c_inb(int minor) 
{
	/* read byte via i2c port, without start/stop sequence	*/
	/* acknowledge is sent in i2c_read.			*/
	int i;
	char data;

	/* assert: scl is low */
	DEB2(printk("i2c_inb.\n"));

	sdahi(minor);
	data=0;
	for (i=0;i<8;i++){
		if (sclhi(minor)<0){ /* timeout */
			sdahi(minor); /* we don't want to block the net */
			return 0;
		};
		data*=2;
		if ( sdain(minor) ) data++;
		scllo(minor);
	}
	/* assert: scl is low */
	return data;
}


/* ----- level 2: communication with the kernel  ----- */

static int i2c_open(struct inode * inode, struct file * file) 
{
	return 0;
}


static void i2c_release (struct inode * inode, struct file * file) 
{
}


static int i2c_write(struct inode * inode, struct file * file,
	const char * buf, int count)
{
	unsigned int minor = MINOR(inode->i_rdev);
	char c;
	char adr;
	const char *temp = buf;
	int retval,i;

	i2c_start(minor);
	i=0;
	while (!i2c_outb(minor,get_fs_byte(buf)&0xfe)) {
		i2c_stop(minor);
		i++;
		i2c_start(minor);
		if (i>=i2c_table[minor].retries ) {
			DEBACK(printk("i2c_read: i2c%d: address ack failed.\n",minor));
			i2c_stop(minor);
			return -EREMOTEIO;
		};
	}
	count--;
	temp++;
	
	while (count > 0) {
		c = get_fs_byte(temp);

		DEB2(printk("i2c(%d)_write: writing %c\n",minor,c));

		retval = i2c_outb(minor,c);
		if (retval>0) { 
			count--; 
			temp++;
		} else { /* arbitration or no acknowledge */
			DEBE(printk("i2c(%d)_write: error - bailout.\n",minor));
			i2c_stop(minor);
			return -EREMOTEIO; /* got a better one ?? */
		}
		__delay(i2c_table[minor].mdelay * (loops_per_sec / 1000) );
	}
	i2c_stop(minor);
	DEB2(printk("i2c_write: wrote %d bytes.\n",temp-buf));
	return temp-buf ;
 }

static int i2c_read(struct inode * inode, struct file * file,
	char * buf, int count) 
{
	unsigned int minor = MINOR(inode->i_rdev);
	char *temp = buf;
	char adr;
	int i;

	i2c_start(minor);
	i=0;
	while (!i2c_outb(minor,get_fs_byte(buf)|0x01)) {
		i2c_stop(minor);
		i++;
		if (i>=i2c_table[minor].retries )  {
			DEBACK(printk("i2c_read: i2c%d: address ack failed.\n",minor));
			i2c_stop(minor);
			return -EREMOTEIO;
		};
		i2c_start(minor);
	}

	while (count > 0) {
		put_user(i2c_inb(minor),temp);

		if (count==1){		/* send ack	*/
			sdahi(minor);
		}else{
			sdalo(minor);		/* reverse ack on last byte */
		}
		if (sclhi(minor)<0) { /* timeout */
			sdahi(minor);
			return 0;
		};
		scllo(minor);
		sdahi(minor);
		temp++;
		count--;
	}

	i2c_stop(minor);
	DEB(printk("i2c_read: %d bytes read.\n",temp-buf));
	return temp-buf ;
}

static int i2c_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	unsigned int minor = MINOR(inode->i_rdev);
	int retval = 0;

	switch ( cmd ) {
		case I2C_TIMEOUT:
			i2c_table[minor].timeout = arg;
			break;
		case I2C_UDELAY:
			i2c_table[minor].udelay = arg;
			case I2C_MDELAY:
	 			i2c_table[minor].mdelay = arg;
			break;
		case I2C_RETRIES:
			i2c_table[minor].retries = arg;
			break;
#ifdef SLO_IO
		case I2C_V_SLOW: 
			i2c_table[minor].veryslow = arg;
			break;
#endif
		case I2C_RESET:
			i2c_reset(minor);
			break;
		case I2C_CLEAR:
			i2c_table[minor].flags = P_EXIST |ADAP_PHIL;
			break;
		default:
			retval = -EINVAL;
	}
	return retval;
}

static int i2c_init(int minor)
{	
	if (check_region(i2c_table[minor].base, 
		(i2c_table[minor].base == 0x3bc)? 3 : 8) < 0 ) {
		DEBE(printk("i2c_init: Port %#x already in use.\n",
			i2c_table[minor].base));
		return -ENODEV;
	} else {
		unsigned char tst;
		/* now do some simple line checking			*/
		/* assert: we are alone on the bus 			*/
		tst=inb(STAT);
		if ( !(tst & I2C_ON) ){
			printk("i2c%d: power is down\n",minor);
		}
		scllo(minor);
		if ( ( inb(STAT) & I2C_SCLIN ) ) {
			printk("i2c%d SCL line stuck high!\n",minor);
			return -EIO;
		}
		sclhi(minor);
		sdalo(minor);	
		if ( inb(STAT) & I2C_SDAIN ) {
			printk("i2c%d SDA line stuck high!\n",minor);
			return -EIO;			
		}
		sdahi(minor);	
		if ( ! ( inb(STAT) & I2C_SDAIN ) ) {
			printk("i2c%d SDA line stuck low!\n",minor);
			return -EIO;			
		}
		request_region(i2c_table[minor].base, 
			(i2c_table[minor].base == 0x3bc)? 3 : 8, 
			"i2c (Philips adapter)");
		i2c_reset(minor);
	}
	return 0;
}

static void i2c_exit(int minor)
{	
	release_region( i2c_table[minor].base , 
		(i2c_table[minor].base == 0x3bc)? 3 : 8 );
}

/* -----exported file operations: -------------------------------------	*/
struct i2c_opns lpbitops = {
	i2c_read,
	i2c_write,
	i2c_ioctl,
	i2c_open,
	i2c_release,
	i2c_init,
	i2c_exit
};
